
   /**-------------------------------------------------------------------**
    **                               CLooG                               **
    **-------------------------------------------------------------------**
    **                             domain.h                              **
    **-------------------------------------------------------------------**
    **                  First version: october 28th 2001                 **
    **-------------------------------------------------------------------**/


/******************************************************************************
 *               CLooG : the Chunky Loop Generator (experimental)             *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2001 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CLooG, the Chunky Loop Generator                                           *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/


#ifndef CLOOG_DOMAIN_H
#define CLOOG_DOMAIN_H
#if defined(__cplusplus)
extern "C" 
  {
#endif 


# define MAX_RAYS 200


struct cloogdomainlist
{ Polyhedron * domain ;
  struct cloogdomainlist * next ;
} ;
typedef struct cloogdomainlist CloogDomainList ;


/******************************************************************************
 *                              PolyLib interface                             *
 ******************************************************************************/
void         cloog_domain_matrix_print(FILE *, Matrix *) ;
void         cloog_domain_matrix_free(Matrix *) ;
Matrix     * cloog_domain_matrix_alloc(unsigned, unsigned) ;
Polyhedron * cloog_domain_matrix2domain(Matrix *) ;

void         cloog_domain_print(FILE *, Polyhedron *) ;
void         cloog_domain_free(Polyhedron *) ;
Matrix     * cloog_domain_domain2matrix(Polyhedron *) ;
Polyhedron * cloog_domain_copy(Polyhedron *) ;
Polyhedron * cloog_domain_image(Polyhedron *, Matrix *) ;
Polyhedron * cloog_domain_preimage(Polyhedron *, Matrix *) ;
Polyhedron * cloog_domain_convex(Polyhedron * Pol) ;
Polyhedron * cloog_domain_simplify(Polyhedron *, Polyhedron *) ;
Polyhedron * cloog_domain_union(Polyhedron *, Polyhedron *) ;
Polyhedron * cloog_domain_intersection(Polyhedron *, Polyhedron *) ;
Polyhedron * cloog_domain_difference(Polyhedron *, Polyhedron *) ;
int          cloog_domain_includes(Polyhedron *, Polyhedron *) ;
Polyhedron * cloog_domain_addconstraints(Polyhedron *, Polyhedron *) ;
void         cloog_domain_sort(Polyhedron **,unsigned,unsigned,unsigned,int *) ;


/******************************************************************************
 *                          Structure display function                        *
 ******************************************************************************/
void cloog_domain_list_print(FILE *, CloogDomainList *) ;


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/
void cloog_domain_list_free(CloogDomainList *) ;


/******************************************************************************
 *                               Reading function                             *
 ******************************************************************************/
Polyhedron * cloog_domain_read(FILE *) ;
Polyhedron * cloog_domain_union_read(FILE *) ;
CloogDomainList * cloog_domain_list_read(FILE *) ;


/******************************************************************************
 *                            Processing functions                            *
 ******************************************************************************/
Polyhedron * cloog_domain_compact(Polyhedron *) ;
int cloog_domain_empty(Polyhedron *) ;
int cloog_domain_universe(Polyhedron *) ;
Polyhedron * cloog_domain_project(Polyhedron *, int, int) ;
Polyhedron * cloog_domain_extend(Polyhedron *, int, int) ;
int cloog_domain_never_integral(Polyhedron *, int, int) ;
int cloog_domain_gcd(int, int) ;
Value cloog_domain_stride(Polyhedron *, int, int, Value *) ;
int cloog_domain_integral_lowerbound(Polyhedron *, int, Value *) ;
void cloog_domain_lowerbound_update(Polyhedron *, int, Value) ;
int cloog_domain_prism_level(Polyhedron *, int) ;
int cloog_domain_point(Polyhedron *, int, int) ;
int cloog_domain_quick_point(Polyhedron *, int) ;
int cloog_domain_quick_zero(Polyhedron *, int) ;
int cloog_domain_quick_equal(Polyhedron *, Polyhedron *) ;
int cloog_domain_quick_disjoint(Polyhedron *, Polyhedron *) ;
int cloog_domain_list_quick_same(CloogDomainList *) ;
Polyhedron * cloog_domain_grow(Polyhedron *, int, int) ;
Polyhedron * cloog_domain_restrict(Polyhedron *, Polyhedron *, int) ;

#if defined(__cplusplus)
  }
#endif 
#endif /* define _H */
