
   /**-------------------------------------------------------------------**
    **                              CLooG                                **
    **-------------------------------------------------------------------**
    **                             loop.h                                **
    **-------------------------------------------------------------------**
    **                 First version: october 26th 2001                  **
    **-------------------------------------------------------------------**/


/******************************************************************************
 *               CLooG : the Chunky Loop Generator (experimental)             *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2001 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CLooG, the Chunky Loop Generator                                           *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/


#ifndef CLOOG_LOOP_H
#define CLOOG_LOOP_H
#if defined(__cplusplus)
extern "C" 
  {
#endif 


struct cloogloop
{ Polyhedron * domain ;        /* The iteration domain. */
  int stride ;                 /* The stride for the corresponding iterator
                                * (filled only after loop generation).
				*/
  CloogStatement * statement ; /* The included statements. */
  struct cloogloop * inner ;   /* Loops at the next level. */
  struct cloogloop * next ;    /* Next loop at the same level. */
} ;
typedef struct cloogloop CloogLoop ;


/******************************************************************************
 *                          Structure display function                        *
 ******************************************************************************/
void cloog_loop_print(FILE *, CloogLoop *) ;


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/
void cloog_loop_free(CloogLoop *) ;


/******************************************************************************
 *                              Reading functions                             *
 ******************************************************************************/
CloogLoop * cloog_loop_read(FILE *, int, int) ;


/******************************************************************************
 *                            Processing functions                            *
 ******************************************************************************/
CloogLoop * cloog_loop_generate(CloogLoop *, Polyhedron *, int, int,
                                CloogOptions *) ;
CloogLoop * cloog_loop_simplify(CloogLoop *, Polyhedron *, int, int) ;
CloogLoop * cloog_loop_fast_simplify(CloogLoop *, Polyhedron *, int, int) ;
void cloog_loop_scatter(CloogLoop *, Polyhedron *) ;
void cloog_loop_strides(CloogLoop *, int) ;


#if defined(__cplusplus)
  }
#endif 
#endif /* define _H */
