
   /**-------------------------------------------------------------------**
    **                              CLooG                                **
    **-------------------------------------------------------------------**
    **                             names.c                               **
    **-------------------------------------------------------------------**
    **                  First version: august 1st 2002                   **
    **-------------------------------------------------------------------**/


/******************************************************************************
 *               CLooG : the Chunky Loop Generator (experimental)             *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2002 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CLooG, the Chunky Loop Generator                                           *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/
/* CAUTION: the english used for comments is probably the worst you ever read,
 *          please feel free to correct and improve it !
 */


# include <stdlib.h>
# include <stdio.h>
# include <ctype.h>
# include "../include/cloog/cloog.h"


/******************************************************************************
 *                          Structure display function                        *
 ******************************************************************************/


/* cloog_names_print function:
 * This function prints the content of a CloogNames structure (names) into a
 * file (foo, possibly stdout).
 */
void cloog_names_print(FILE * foo, CloogNames * names)
{ int i ;
  
  if (names != NULL)
  { if (names->nb_parameters > 0)
    { fprintf(foo,"Parameters are:\n") ;
      for (i=0;i<names->nb_parameters;i++)
      fprintf(foo,"%s ",names->parameters[i]) ;
      fprintf(foo,"\n") ;
    }
    else
    fprintf(foo,"There is no parameters.\n") ;
    
    fprintf(foo,"Iterators are:\n") ;
    for (i=0;i<names->nb_iterators;i++)
    fprintf(foo,"%s ",names->iterators[i]) ;
    fprintf(foo,"\n") ;
  }
  else
  fprintf(foo,"There are no name setting.\n") ;
}


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/


/* cloog_names_free function:
 * This function frees the allocated memory for a CloogNames structure.
 */
void cloog_names_free(CloogNames * names)
{ int i ;

  if (names->iterators != NULL)
  { for (i=0;i<names->nb_iterators;i++)
    free(names->iterators[i]) ;
    free(names->iterators) ;
  }
   
  if (names->parameters != NULL)
  { for (i=0;i<names->nb_parameters;i++)
    free(names->parameters[i]) ;
    free(names->parameters) ;
  }
  free(names) ;
}


/******************************************************************************
 *                              Reading functions                             *
 ******************************************************************************/


/* cloog_names_read function:
 * This function reads names data from a file (foo, possibly stdin). It first
 * reads the naming option to know if whether has to automatically generate the
 * names, or to read them. Names are stored into an array of strings, and a
 * pointer to this array is returned.
 * - nb_items is the number of names the function will have to read if the
 *   naming option is set to read.
 * - prefix is the prefix to give to each name in case of automatic generation.
 * - first item is the name of the first suffix in case of automatic generation.
 * September 9th 2002: first version.
 */
char ** cloog_names_read(foo, nb_items, prefix, first_item)
FILE * foo ;
int nb_items ;
char * prefix, first_item ;
{ int i, option, n ;
  char s[MAX_STRING], str[MAX_STRING], * c, ** names ;

  /* We first read name option. */
  while (fgets(s,MAX_STRING,foo) == 0) ;
  while ((*s=='#' || *s=='\n') || (sscanf(s," %d",&option)<1))
  fgets(s,MAX_STRING,foo) ;
  
  /* If there is no item to read, then return NULL. */
  if (nb_items == 0)
  return NULL ;
  
  /* If option is to read them in the file, then we do it and put them into
   * the array.
   */
  if (option)
  { /* Memory allocation. */
    names = (char **)malloc(nb_items*sizeof(char *)) ;
    if (names == NULL) 
    { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
      exit(1) ;
    }
    for (i=0;i<nb_items;i++)
    { names[i] = (char *)malloc(MAX_NAME*sizeof(char)) ;
      if (names[i] == NULL) 
      { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
        exit(1) ;
      }
    }
    
    do  /* Skip the comments, spaces and empty lines... */
    { c = fgets(s,MAX_STRING,foo) ;
      while ((c != NULL) && isspace(*c) && (*c != '\n'))
      c++ ;
    }
    while (c != NULL && (*c == '#' || *c == '\n'));
    
    if (c == NULL) 
    { fprintf(stderr, "[CLooG]ERROR: no names in input file.\n") ;
      exit(1) ;
    }
    for (i=0;i<nb_items;i++) 
    { /* All names must be on the same line. */
      while (isspace(*c))
      c++ ;
      if (c == NULL || *c == '#' || *c == '\n')
      { fprintf(stderr, "[CLooG]ERROR: not enough names in input file.\n") ;
        exit(1) ;
      }
      /* n is strlen(str). */
      if (sscanf(c,"%s%n",str,&n) == 0) 
      { fprintf(stderr, "[CLooG]ERROR: no names in input file.\n") ;
        exit(1) ;
      }
      sscanf(str,"%s",names[i]) ;
      c += n ;
    }
  }
  /* Else we create names automatically. */
  else
  names = cloog_names_generate_items(nb_items,prefix,first_item) ;

  return names ;
}


/******************************************************************************
 *                            Processing functions                            *
 ******************************************************************************/


/* cloog_names_generate_items function:
 * This function returns a pointer to an array of strings with entries set
 * thanks to the function's parameters.
 * - nb_items will be the number of entries in the string array.
 * - prefix is the name prefix of each item.
 * - first_item is the name of the first item (without prefix), the nb_items-1
 *   following items will be the nb_items-1 following letters in ASCII code.
 * September 9th 2002 : first version, extracted from cloog_names_generate.
 */
char ** cloog_names_generate_items(int nb_items, char * prefix, char first_item)
{ int i ;
  char ** names ;
    
  names = (char **)malloc(nb_items*sizeof(char *)) ;
  if (names == NULL) 
  { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
    exit(1) ;
  }
  for (i=0;i<nb_items;i++)
  { names[i] = (char *)malloc(MAX_NAME*sizeof(char)) ;
    if (names[i] == NULL) 
    { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
      exit(1) ;
    }
    if (prefix == NULL)
    sprintf(names[i],"%c",first_item+i) ;
    else
    sprintf(names[i],"%s%c",prefix,first_item+i) ;
  }
  
  return names ;
}


/* cloog_names_generate function:
 * This function returns a pointer to a CloogNames structure with fields set
 * thanks to the function's parameters.
 * - nb_iterators will be the number of iterators in the CloogNames structure.
 * - nb_parameters will be the number of parameters in the CloogNames structure.
 * - first_i is the name of the first iterator, the nb_iterators-1 following
 *   iterators will be the nb_iterators-1 following letters in ASCII code.
 * - first_p is the name of the first parameter, the nb_parameters-1 following
 *   parameters will be the nb_parameters-1 following letters in ASCII code.
 * July 1st 2002 : first version.
 * September 9th 2002 : use of cloog_names_generate_items.
 */
CloogNames * cloog_names_generate(nb_iterators, nb_parameters, first_i, first_p)
int nb_iterators, nb_parameters ;
char first_i, first_p ;
{ CloogNames * names ;

  names = (CloogNames *)malloc(sizeof(CloogNames)) ;
  if (names == NULL) 
  { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
    exit(1) ;
  }
  
  names->nb_parameters = nb_parameters ;
  names->nb_iterators  = nb_iterators ;
  names->parameters = cloog_names_generate_items(nb_parameters,NULL,first_p) ;
  names->iterators  = cloog_names_generate_items(nb_iterators,NULL,first_i) ;

  return names ;
}

