
   /**-------------------------------------------------------------------**
    **                              CLooG                                **
    **-------------------------------------------------------------------**
    **                             pprint.c                              **
    **-------------------------------------------------------------------**
    **                  First version: october 26th 2001                 **
    **-------------------------------------------------------------------**/


/******************************************************************************
 *               CLooG : the Chunky Loop Generator (experimental)             *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2001 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CLooG, the Chunky Loop Generator                                           *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/
/* CAUTION: the english used for comments is probably the worst you ever read,
 *          please feel free to correct and improve it !
 */

# include <stdlib.h>
# include <stdio.h>
# include <string.h>
# include "../include/cloog/cloog.h"


/******************************************************************************
 *                        Equalities spreading functions                      *
 ******************************************************************************/


/* pprint_equal_print function:
 * This function prints the content of a CloogEqual structure (equal) into a
 * file (foo, possibly stdout).
 */
void pprint_equal_print(FILE * foo, CloogEqual * equal)
{ fprintf(foo,"Equal:\n") ;
  
  if (equal == NULL)
  fprintf(foo,"Null equal.\n") ;
  else
  { while (equal != NULL)
    { fprintf(foo,"Type %d, at line %d ",equal->type,equal->line) ;
      fprintf(foo,"for level %d:\n",equal->level) ;
      cloog_domain_print(foo,equal->domain) ;
      fprintf(foo,"\n") ;
      
      equal = equal->next ;
    }     
  }
}


/* pprint_equal function:
 * This function returns the content of a linked list (equal) into a string.
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * July   2nd 2002: first version. 
 * March 16th 2003: return now a string instead of printing directly and do not
 *                  write 'Sx()' if there is no spreading, but only 'Sx'. 
 */
char * pprint_equal(equal, infos)
CloogEqual * equal ;
CloogInfos * infos ;
{ int first=1 ;
  char * body, * temp, * seq, * sline ;
  CloogEqual * copy ;

  copy = equal ;
  temp = (char *)malloc(MAX_STRING*sizeof(char)) ;
  body = (char *)malloc(MAX_STRING*sizeof(char)) ;
  body[0] = '\0' ;
  seq = (char *)malloc(MAX_STRING*sizeof(char)) ;
      
  while (equal != NULL)
  { /* It is not necessary to print here the scattering iterators since they
     * never appear in the statement bodies.
     */
    if (equal->level > infos->scattdims)
    { if (!first)
      strcat(body,",") ;
      
      sprintf(temp,"%s = ",infos->iterators[equal->level-1]) ;
      strcat(body,temp) ;
    
      if (equal->line != ONE_TIME_LOOP)
      sline = pprint_line(equal->domain,copy,equal->line,equal->level,infos) ;
      else
      sline = pprint_minmax(equal->domain,copy,equal->level,1,0,infos) ;
    
      strcat(body,sline) ;    
      free(sline) ;
      first = 0 ;
    }
    equal = equal->next ;
  }
  free(temp) ;

  if (!first)
  sprintf(seq,"(%s)",body) ;
  else
  sprintf(seq,"%s",body) ;

  free(body) ;

  return(seq) ;
}


/* pprint_equal_cpp function:
 * This function prints the substitution data of a statement into a string.
 * Using this function instead of pprint_equal is useful for generating
 * a compilable pseudo-code by using preprocessor macro for each statement.
 * By opposition to pprint_equal, the result is less human-readable. For
 * instance this function will print (i,i+3,k,3) where pprint_equal would
 * return (j=i+3,l=3).
 * - level is the number of loops enclosing the statement,
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * March 12th 2004: first version. 
 */
char * pprint_equal_cpp(equal, level, infos)
CloogEqual * equal ;
int level ;
CloogInfos * infos ;
{ int i ;
  char * temp, * seq, * sline ;
  CloogEqual * copy ;

  copy = equal ;
  temp = (char *)malloc(MAX_STRING*sizeof(char)) ;
  seq = (char *)malloc(MAX_STRING*sizeof(char)) ;
  seq[0] = '\0' ;
  
  /*printf("scattdims = %d, level-1 = %d\n",infos->scattdims,level-1) ;
  pprint_equal_print(stdout,equal) ;*/
  strcat(seq,"(") ;
  for (i=infos->scattdims;i<level-1;i++)
  { while (equal != NULL)
    { if (equal->level < i+1)
      equal = equal->next ;
      else
      break ;
    }
  
    if ((equal != NULL) && (equal->level == i+1))
    { if (equal->line != ONE_TIME_LOOP)
      sline = pprint_line(equal->domain,copy,equal->line,equal->level,infos) ;
      else
      sline = pprint_minmax(equal->domain,copy,equal->level,1,0,infos) ;
      
      strcat(seq,sline) ;    
      free(sline) ;
      
      equal = equal->next ;
    }
    else
    { sprintf(temp,"%s",infos->iterators[i]) ;
      strcat(seq,temp) ;
    }
    
    if (i != level-2)
    strcat(seq,",") ;
  }
  free(temp) ;
  strcat(seq,")") ;

  return(seq) ;
}

 
/* pprint_equal_type function :
 * This function returns the type of the equality in the constraint (line) of
 * (domain) for the element (level). An equality is 'constant' iff all other
 * factors are null except the constant one. It is a 'pure item' iff one and
 * only one factor is non null and is 1 or -1. Otherwise it is an 'affine
 * expression'.
 * For instance:
 *   i = -13 is constant, i = j, j = -M are pure items,
 *   j = 2*M, i = j+1 are affine expressions.
 * When the equality comes from a 'one time loop', (line) is ONE_TIME_LOOP.
 * This case require a specific treatment since we have to study all the
 * constraints.
 * - domain is the polyhedron containing all the constraints,
 * - level is the column number in domain of the element which is 'equal to',
 * - line is the line number in domain of the constraint we want to study;
 *   if it is -1, all lines must be studied.
 * July 3rd 2002: first version, called pprint_equal_isconstant. 
 * July 6th 2002: adaptation for the 3 types. 
 */
int pprint_equal_type(domain, level, line)
Polyhedron * domain ;
int level, line ;
{ int i, j, one=0, allnull=0, coef=0, nonnullcst=0 ;
  Value * expr ;

  expr = domain->Constraint[line] ;
  
  if (line != ONE_TIME_LOOP)
  { /* There is only one non null factor, and it must be +1 or -1 for
     * iterators or parameters.
     */ 
    for (i=1;i<=domain->Dimension;i++)
    if ((expr[i] != 0) && (i != level))
    { if (((expr[i] != 1) && (expr[i] != -1)) || (one != 0))
      return EQTYPE_EXAFFINE ;
      else
      one = 1 ;
    }
    /* if the constant factor is non null, it must be alone. */
    if (one != 0)
    { if (expr[domain->Dimension+1] != 0)
      return EQTYPE_EXAFFINE ;
    }
    else
    return EQTYPE_CONSTANT ;
  }
  else
  { /* CAUTION : more tests are necessary for this part to be bug safe. */
    /* We test all lines in the same spirit as before. */
    for (i=0;i<domain->NbConstraints;i++)
    { expr = domain->Constraint[i] ;
      if (expr[level] > 0)
      { for (j=1;j<=domain->Dimension;j++)
        if ((expr[j] != 0) && (j != level))         
        { /* New non nul factor must be 1 or -1, be unique (new or same as
           * before), and if same as before, with same value.
           */
          if (((expr[j] != 1) && (expr[j] != -1)) || 
              ((one != 0) && (one != j)) ||
              ((one != j) && (expr[j] != coef)))
          return EQTYPE_EXAFFINE ;
          else
          { /* A factor must not have been seen at 0 before and non nul after.*/
            if (allnull)
            return EQTYPE_EXAFFINE ;
            one = j ;
            coef = expr[j] ;
          }
        }
        /* We expect Polylib to give only one expression with an item and a
         * constant, for instance : no i - 2 >= 0 AND i - 4 >= 0 !
         */
        if (expr[domain->Dimension+1] != 0)
        { if (one != 0)
          return EQTYPE_EXAFFINE ;
          else
          nonnullcst = 1 ;
        }
        if (one == 0)
        allnull = 1 ;
      }
    }
    if (nonnullcst || (one == 0))
    return EQTYPE_CONSTANT ;
  }  
  return EQTYPE_PUREITEM ;
}


/* pprint_equal_add function:
 * This function adds at the end of the linked list (equal) a new node with the
 * fields domain (domain), level (level) and line (line). It may happen that
 * the created node is the first of the linked list. It returns 1 if the new
 * node can be added, 0 otherwise.
 * - the infos structure gives the user some options about code printing,
 * July 2nd 2002: first version. 
 */
int pprint_equal_add(equal, domain, level, line, infos)
CloogEqual ** equal ;
Polyhedron * domain ;
int level, line ;
CloogInfos * infos ;
{ CloogEqual * new, * temp ;
  int type ;
  
  /* First, we check whether the options allow us to spread the equality. */
  if ((!infos->options->csp && !infos->options->esp) ||
      (level < infos->options->fsp))
  return 0 ;
  type = pprint_equal_type(domain,level,line) ;
  if (infos->options->csp && (type == EQTYPE_EXAFFINE) && !infos->options->esp)
  return 0 ;

  /* If yes, we add it to the linked list. */
  if (*equal == NULL)
  { *equal = (CloogEqual *)malloc(sizeof(CloogEqual)) ;
    if (*equal == NULL)
    { fprintf(stderr, "Memory Overflow.\n") ;
      exit(1) ;
    }
    (*equal)->domain = domain ;
    (*equal)->line = line ;
    (*equal)->level = level ;
    (*equal)->type = type ;
    (*equal)->next = NULL ;
  }
  else
  { temp = *equal ;
    while (temp->next != NULL)
    temp = temp->next ;
    
    new = (CloogEqual *)malloc(sizeof(CloogEqual)) ;
    if (new == NULL)
    { fprintf(stderr, "Memory Overflow.\n") ;
      exit(1) ;
    }
    new->domain = domain ;
    new->line = line ;
    new->level = level ;
    new->type = type ;
    new->next = NULL ;
    temp->next = new ;
  }
  return 1 ;
}
  

/* pprint_equal_del function :
 * This function removes and frees the last node of the linked list (equal) if
 * its field level is (level).
 * July 2nd 2002: first version. 
 */
void pprint_equal_del(CloogEqual ** equal, int level)
{ CloogEqual * prec, * temp ;

  if (*equal != NULL)
  { if ((*equal)->next == NULL)
    { if ((*equal)->level == level)
      { free(*equal) ;
        *equal = NULL ;
      }
    }
    else
    { temp = *equal ;
      while  (temp->next != NULL)
      { prec = temp ;
        temp = temp->next ;
      }
      if (temp->level == level)
      { free(prec->next) ;
        prec->next = NULL ;
      }
    }
  }
}


/******************************************************************************
 *                       Pretty Printing (dirty) functions                    *
 ******************************************************************************/


/* pprint_val function:
 * This function returns a string containing the printing of a value (possibly
 * an iterator or a parameter with its coefficient or a constant).
 * - equal is a linked list containing the equalities to spread,
 * - val is the coefficient or constant value,
 * - first is a pointer to a boolean set to 1 if the current value is the first
 *   of an expresion, 0 otherwise (this function can change it),
 * - cst is a boolean set to 1 if the value is a constant, 0 otherwise,
 * - name is a string containing the name of the iterator or of the parameter,
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * November 2nd 2001: first version. 
 * November 8th 2001: complete rewriting. 
 * July 6th 2002: integration of the equality spreading.
 * October 10th 2002: (debug) sline was sometimes freed twice (hard to find !),
 *                    and equality spreading gives no more (for instance) --2
 *                    but -(-2).
 * June 27th 2003: 64 bits version ready.
 */
char * pprint_val(equal, level, val, first, cst, name, infos)
CloogEqual * equal ;
Value val ;
int level, * first, cst ;
char * name ;
CloogInfos * infos ;
{ char * sval, * body, * sline ;

  body = (char *)malloc(MAX_STRING_VAL*sizeof(char)) ;
  sval = (char *)malloc(MAX_STRING_VAL*sizeof(char)) ;
  body[0] = '\0' ;
  sval[0] = '\0' ;

  /* statements for the spreading of equalities. */
  while (equal != NULL)   
  { if (equal->level == level)
    { if (equal->line != ONE_TIME_LOOP)
      sline = pprint_line(equal->domain,NULL,equal->line,equal->level,infos) ;
      else
      sline = pprint_minmax(equal->domain,NULL,equal->level,1,0,infos) ;

      if (((equal->type == EQTYPE_EXAFFINE) && (val != 1)) ||
          (val < 0))
      { sprintf(body,"(") ;
        strcat(body,sline) ;
        strcat(body,")") ;
      }
      else
      sprintf(body,"%s",sline) ;
      free(sline) ;
      name = body ;             
      break ;
    }      
    equal = equal->next ;     
  }

  /* statements for the 'normal' processing. */
  if ((val != 0) && (!cst))
  { if ((*first) || (val < 0))
    { switch (val)
      { case  1 : sprintf(sval,"%s",name) ;
                  break ;
        case -1 : sprintf(sval,"-%s",name) ;
                  break ;
        default : sprintf(sval,VALUE_FMT"*%s",val,name) ;
      }
      *first = 0 ;
    }
    else
    { if (val == 1)
      sprintf(sval,"+%s",name) ;
      else
      sprintf(sval,"+"VALUE_FMT"*%s",val,name) ;
    }
  }
  else
  { if (cst)
    { if (((val == 0) && (*first)) || (val < 0))
      sprintf(sval,VALUE_FMT,val) ;
      if (val > 0)
      { if (!(*first))
        sprintf(sval,"+"VALUE_FMT,val) ;
        else
        sprintf(sval,VALUE_FMT,val) ;
      }
    }
  }
  free(body) ;
  
  return(sval) ;
}
 

/* pprint_line function:
 * This function returns a string containing the printing of the 'right part'
 * of a constraint according to an element.
 * For instance, for the constraint -3*i + 2*j - M >=0 and the element j,
 * we have j >= (3*i + M)/2. As we are looking for integral solutions, this
 * function should return 'ceild(3*i+M,2)'.
 * - domain is the polyhedron containing all the constraints,
 * - equal is a linked list containing the equalities to spread,
 * - line_num is the line number in domain of the constraint we want to print,
 * - level is the column number in domain of the element we want to use,
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * November 2nd 2001: first version. 
 * June 27th 2003: 64 bits version ready.
 */
char * pprint_line(domain, equal, line_num, level, infos)
Polyhedron * domain ;
CloogEqual * equal ;
int line_num, level ;
CloogInfos * infos ;
{ int i, nb_iter, sign, first=1, nb_elts=0 ;
  char * sval, * body, * sline, * name ;
  Value * line, numerator, denominator ;

  line = domain->Constraint[line_num] ;
  
  body = (char *)malloc(MAX_STRING*sizeof(char)) ;
  body[0] = '\0' ;
  sline = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sline[0] = '\0' ;

  if (line[level] != 0)
  { /* Maybe we need to invert signs in such a way that the element sign is>0.*/
    sign = (line[level] > 0) ? -1 : 1 ;

    /* First, we have to print the iterators. */
    nb_iter = domain->Dimension - infos->nb_parameters ;
    for (i=1;i<=nb_iter;i++)
    if ((i != level) && (line[i] != 0))
    { name = infos->iterators[i-1] ;
      sval = pprint_val(equal,i,sign*line[i],&first,0,name,infos) ;
      strcat(body,sval) ;
      free(sval) ;
      nb_elts ++ ;
    }    

    /* Next, the parameters. */
    for (i=nb_iter+1;i<=domain->Dimension;i++)
    if ((i != level) && (line[i] != 0))
    { name = infos->parameters[i-nb_iter-1] ;
      sval=pprint_val(NULL,-1,sign*line[i],&first,0,name,infos) ;
      strcat(body,sval) ;
      free(sval) ;
      nb_elts ++ ;
    }    

    numerator = sign * line[domain->Dimension + 1] ;
    denominator = -1 * sign * line[level] ;
        
    /* Finally, the constant, and the final printing. */
    if (!first)
    { sval = pprint_val(NULL,-1,numerator,&first,1,"",infos) ;
      strcat(body,sval) ;
      if (strlen(sval)>0)
      nb_elts ++ ;
      free(sval) ;
    
      if ((line[level] != 1) && (line[level] != -1))
      { if (line[0] == 1)
        { if (line[level] > 0)
          { if (infos->language == LANGUAGE_FORTRAN)
            sprintf(sline,"CEILING(REAL(%s)/REAL("VALUE_FMT"))",body,
                    denominator) ;
            else
            sprintf(sline,"ceild(%s,"VALUE_FMT")",body,denominator) ;
          }
          else
          { if (infos->language == LANGUAGE_FORTRAN)
            sprintf(sline,"FLOOR(REAL(%s)/REAL("VALUE_FMT"))",body,denominator);
            else
            sprintf(sline,"floord(%s,"VALUE_FMT")",body,denominator) ;
          }
        }
        else
        { if (nb_elts > 1)
          sprintf(sline,"(%s)/"VALUE_FMT,body,denominator) ;
          else
          sprintf(sline,"%s/"VALUE_FMT,body,denominator) ;
        }  
      }
      else
      sprintf(sline,"%s",body) ;
    }
    else
    { if (numerator == 0)
      sprintf(sline,"0") ;
      else
      { if (denominator != 1)
        { if (line[0] == 1) /* useful? */
          { if (numerator%denominator == 0)
            sprintf(sline,VALUE_FMT,numerator/denominator) ;
            else
            { if (numerator < 0)
              { if (line[level] > 0)
	        /* nb<0 need max */
                sprintf(sline,VALUE_FMT,numerator/denominator) ;
                else
		/* nb<0 need min */
                sprintf(sline,VALUE_FMT,numerator/denominator-1) ;
              }
              else
              { if (line[level] > 0)
	        /* nb>0 need max */
                sprintf(sline,VALUE_FMT,numerator/denominator+1) ;
                else
		/* nb>0 need min */
                sprintf(sline,VALUE_FMT,numerator/denominator) ;
              }
            }
          }
          else
          sprintf(sline,VALUE_FMT"/"VALUE_FMT,numerator,denominator) ;
        }
        else
        sprintf(sline,VALUE_FMT,numerator) ;      
      }
    }
  }
  free(body) ;
  
  return(sline) ;
}


/* pprint_minmax function:
 * This function returns a string containing the printing of a minimum or a
 * maximum of the 'right parts' of all constraints according to an element.
 * For instance consider the constraints:
 * -3*i  +2*j   -M >= 0
 *  2*i    +j      >= 0
 *   -i    -j +2*M >= 0
 * if we are looking for the minimum for the element j, the function should
 * return 'max(ceild(3*i+M,2),-2*i)'.
 * - domain is the polyhedron containing all the constraints,
 * - equal is a linked list containing the equalities to spread,
 * - level is the column number in domain of the element we want to use,
 * - max is a boolean set to 1 if we are looking for a maximum, 0 for a minimum,
 * - guard is set to 0 if there is no guard, and set to the level of the element
 *   with a guard otherwise (then the function gives the max or the min only
 *   for the constraint where the guarded coefficient is 0), 
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * November 2nd 2001: first version. 
 */
char * pprint_minmax(domain, equal, level, max, guard, infos)
Polyhedron * domain ;
CloogEqual * equal ;
int level, max, guard ;
CloogInfos * infos ;
{ int i, first=1 ;
  char * sline, * body, * sminmax ;
  
  body = (char *)malloc(MAX_STRING*sizeof(char)) ;
  body[0] = '\0' ;
  sminmax = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sminmax[0] = '\0' ;

  for (i=0;i<domain->NbConstraints;i++)
  if (((max && domain->Constraint[i][level] > 0) ||
       (!max && domain->Constraint[i][level] < 0)) &&
      (!guard || (domain->Constraint[i][guard] == 0)) &&
      (domain->Constraint[i][0]))
  { sline = pprint_line(domain,equal,i,level,infos) ;
    strcat(sminmax,sline) ;
    free(sline) ;
    break ;
  }
  /* We begin after the first found element. */
  for (i++;i<domain->NbConstraints;i++)
  if (((max && domain->Constraint[i][level] > 0) ||
       (!max && domain->Constraint[i][level] < 0)) &&
      (!guard || (domain->Constraint[i][guard] == 0)) &&
      (domain->Constraint[i][0]))
  { sline = pprint_line(domain,equal,i,level,infos) ;
    
    if (infos->language == LANGUAGE_FORTRAN)
    { if (first)
      { if (max)
        sprintf(body,"MAX(%s,%s",sminmax,sline) ;
        else
        sprintf(body,"MIN(%s,%s",sminmax,sline) ;
        
        first = 0 ;
        sprintf(sminmax,"%s",body) ;
      }
      else
      { sprintf(body,",%s",sline) ;
        strcat(sminmax,body) ;
      }
    }
    else
    { if (max)
      sprintf(body,"max(%s,%s)",sminmax,sline) ;
      else
      sprintf(body,"min(%s,%s)",sminmax,sline) ;
  
      sprintf(sminmax,"%s",body) ;
    } 
    
    free(sline) ;
  }
  free(body) ;
  
  if ((infos->language == LANGUAGE_FORTRAN) && (!first))
  strcat(sminmax,")") ;
  
  return(sminmax) ;
}


/* pprint_guard function:
 * This function returns a string containing the printing of a guard.
 * A guard on an element (level) is :
 * -> the conjunction of all the existing constraints where the coefficient of
 *    this element is 0 if the element is an iterator,
 * -> the conjunction of all the existing constraints if the element isn't an
 *    iterator.
 * For instance, considering these constraints and the element j:
 * -3*i +2*j -M >= 0
 *  2*i      +M >= 0
 * this function should return 'if (2*i+M>=0) {'.
 * - domain is the polyhedron containing all the constraints,
 * - equal is a linked list containing the equalities to spread,
 * - level is the column number of the element in domain we want to use,
 * - guard is a pointer to a boolean set to 0 if there is no guard, 1 otherwise,
 *   this function can set this boolean to 1 if there is a guard,
 * - indent is a pointer to an integer giving the indent level, this function
 *   can increase indent if there is a guard,
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * November 3rd 2001: first version. 
 * November 14th 2001: a lot of 'purifications'. 
 * July 31th 2002 : (debug) some guard parts are no more redundants. 
 * August 12th 2002 : polyhedra union ('or' conditions) are now supported.
 */
char * pprint_guard(domain, equal, level, guard, indent, infos)
Polyhedron * domain ;
CloogEqual * equal ;
int level, * guard, * indent ;
CloogInfos * infos ;
{ int i, j, k, l, guarded, minmax=-1, nb_or=0, nb_and ;
  char * name, * sline, * body, * temp, * sorguard, * sguard ;
  Polyhedron * copy ;

  temp = (char *)malloc(MAX_STRING*sizeof(char)) ;
  body = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sorguard = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sorguard[0] = '\0' ;
  sguard = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sguard[0] = '\0' ;
  
  while (domain != NULL)
  { copy = Polyhedron_Copy(domain) ;
    
    if (nb_or)
    { if (infos->language == LANGUAGE_FORTRAN)
      strcat(sorguard," .OR. ") ;
      else
      strcat(sorguard," || ") ;
    }
      
    nb_and = 0 ;
    body[0] = '\0' ;
    /* We search for guard parts. */
    for (i=1;i<=copy->Dimension;i++)
    for (j=0;j<copy->NbConstraints;j++)
    if ((copy->Constraint[j][i] != 0) &&
        ((copy->Constraint[j][level] == 0) ||
         (copy->Dimension-infos->nb_parameters < level)))
    { if (i <= copy->Dimension-infos->nb_parameters)
      name = infos->iterators[i-1] ;
      else
      name = infos->parameters[i-(copy->Dimension-infos->nb_parameters+1)] ;
      
      if (nb_and)
      { if (infos->language == LANGUAGE_FORTRAN)
        strcat(body," .AND. ") ;
        else
        strcat(body," && ") ;
      }
        
      if (copy->Constraint[j][0] == 0)
      { sprintf(temp,"(%s == ",name) ;
        strcat(body,temp) ;
        sline = pprint_line(copy,equal,j,i,infos) ;
      }
      else
      { if (copy->Constraint[j][i] > 0)
        { sprintf(temp,"(%s >= ",name) ;
          minmax = 1 ;
        }
        else
        { sprintf(temp,"(%s <= ",name) ;
          minmax = 0 ;
        }
        strcat(body,temp) ;
      
        guarded = (copy->Dimension-infos->nb_parameters >= level) ? level : 0 ;
        sline = pprint_minmax(copy,equal,i,minmax,guarded,infos) ;
      }
      strcat(body,sline) ;
      free(sline) ;
      strcat(body,")") ;
      nb_and ++ ;
      /* 'elimination' of the current constraint, this avoid to use one
       * constraint more than once. The current line is always eliminated,
       * and the next lines if they are in a min or a max.
       */
      for (k=i;k<=copy->Dimension;k++)
      copy->Constraint[j][k] = 0 ;
    
      if (minmax != -1)
      for (l=j+1;l<=copy->NbConstraints;l++)
      if (((minmax == 1) && (copy->Constraint[l][i] > 0)) ||
          ((minmax == 0) && (copy->Constraint[l][i] < 0)))
      for (k=i;k<=copy->Dimension;k++)
      copy->Constraint[l][k] = 0 ;
    }

    if (nb_and)
    { nb_or ++ ;
      if (nb_and > 1)
      { strcat(sorguard,"(") ;
        strcat(sorguard,body) ;
        strcat(sorguard,")") ;
      }
      else
      strcat(sorguard,body) ;
    }
    
    Polyhedron_Free(copy) ;
    domain = domain->next ; 
  }
  free(temp) ;
  free(body) ;
  
  if (nb_or)
  { for (k=0;k<(*indent);k++)
    strcat(sguard," ") ;
    
    if (infos->language == LANGUAGE_FORTRAN)
    strcat(sguard,"IF ") ;
    else
    strcat(sguard,"if ") ;
    
    if (nb_or > 1)
    { strcat(sguard,"(") ;
      strcat(sguard,sorguard) ;
      strcat(sguard,")") ;
    }
    else
    strcat(sguard,sorguard) ;
    
    if (infos->language == LANGUAGE_FORTRAN)
    strcat(sguard," THEN\n") ;
    else
    strcat(sguard," {\n") ;
    *guard = 1 ;
    *indent += INDENT_STEP ;
  }
  free(sorguard) ;
  
  return(sguard) ;
}
 
  
/* pprint_equality function:
 * This function returns a string containing the printing of an equality 
 * constraint according to an element.
 * An equality can be preceded by a 'modulo guard'.
 * For instance, consider the constraint i -2*j = 0 and the
 * element j: pprint_equality should return 'if(i%2==0) { j = i/2 ;'.
 * - domain is the polyhedron containing all the constraints,
 * - equal is a linked list containing the equalities to spread,
 * - num is the line number of the constraint in domain we want to print,
 * - level is the column number of the element in domain we want to use,
 * - modulo_guard is a pointer to a boolean set to 0 if there is no modulo
 *   guard, 1 otherwise, this function can set this boolean to 1,
 * - indent is a pointer to an integer giving the indent level, this function
 *   can increase indent if there is a modulo guard,
 * - the infos structure gives the user some options about code printing,
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * November 13th 2001: first version.
 * June 26th 2003: simplification of the modulo guards (remove parts such as
 *                 modulo is 0, compare vivien or vivien2 with a previous
 *                 version for an idea).
 * June 29th 2003: non-unit strides support.
 * July 14th 2003: (debug) no more print the constant in the modulo guard when
 *                 it was previously included in a stride calculation.
 */
char * pprint_equality(domain, equal, num, level, modulo_guard, indent, infos)
Polyhedron * domain ;
CloogEqual ** equal ;
int num, level, * modulo_guard, * indent ;
CloogInfos * infos ;
{ int i, sign, first=1, nb_elts=0, nb_iter=0, guarded=0, in_stride=0 ;
  char * body, * seq, * sval, * sline, * name ;
  Value * line, val ;

  body = (char *)malloc(MAX_STRING*sizeof(char)) ;
  body[0] = '\0' ;
  seq = (char *)malloc(MAX_STRING*sizeof(char)) ;
  seq[0] = '\0' ;
    
  line = domain->Constraint[num] ;
  sign = (line[level] > 0) ? -1 : 1 ;
  nb_iter = domain->Dimension - infos->nb_parameters ;

  /* First, the modulo guard : the iterators... */
  for (i=1;i<=nb_iter;i++)
  if ((i != level) && (line[i] != 0) && (line[i]%line[level] != 0))
  { if ((infos->stride[i-1]%line[level]) != 0)
    { val  = sign*line[i] ;
      name = infos->iterators[i-1] ;
      sval = pprint_val(*equal,i,val,&first,0,name,infos) ;
      strcat(body,sval) ;
      free(sval) ;
      nb_elts ++ ;
    }
    else
    /* We need to know if an element of the equality has not to be printed
     * because of a stride that guarantees that this element can be divided by
     * the current coefficient. Because when there is a constant element, it
     * is included in the stride calculation (more exactly in the strided
     * iterator new lower bound: the 'offset') and we have not to print it. */
    if (infos->stride[i-1] != 1)
    in_stride = 1 ;
  }

  /* ...the parameters... */
  for (i=nb_iter+1;i<=domain->Dimension;i++)
  if ((line[i] != 0) && (line[i]%line[level] != 0))
  { val  = sign * line[i] ;
    name = infos->parameters[i-nb_iter-1] ;
    sval = pprint_val(NULL,-1,val,&first,0,name,infos) ;
    strcat(body,sval) ;
    free(sval) ;
    nb_elts ++ ;
  }    

  /* ...the constant. */
  if ((nb_elts || ((line[domain->Dimension+1]%line[level] != 0) &&
       (!in_stride))) && (line[level] != 1) && (line[level] != -1))
  { val  = sign * line[domain->Dimension+1] ;
    sval = pprint_val(NULL,-1,val,&first,1,"",infos) ;
    strcat(body,sval) ;
    if (strlen(sval)>0)
    nb_elts ++ ;
    free(sval) ;
  
    for (i=0;i<(*indent);i++)
    strcat(seq," ") ;
    
    if (infos->language == LANGUAGE_FORTRAN)
    strcat(seq,"IF (") ;
    else
    strcat(seq,"if (") ;
    
    val = -1 * sign * line[level] ;
    if (infos->language == LANGUAGE_FORTRAN)
    { strcat(seq,"MOD(") ;
      strcat(seq,body) ;
      sprintf(body,","VALUE_FMT") == 0) THEN\n",val) ;
    }  
    else
    { if (nb_elts > 1)
      { strcat(seq,"(") ;
        strcat(seq,body) ;
        strcat(seq,")") ;
      }
      else
      strcat(seq,body) ;
    
      strcat(seq,"\%") ;
      sprintf(body,VALUE_FMT" == 0) {\n",val) ;
    }
        
    strcat(seq,body) ;
    *modulo_guard = 1 ;
    *indent += INDENT_STEP ;
    guarded = 1 ;
  }       
 
  if (!pprint_equal_add(equal,domain,level,num,infos))
  { /* Finally, the equality. */
    for (i=0;i<(*indent);i++)
    strcat(seq," ") ;
		
    /* If there is no modulo guard and we have to make a block by dimension. */
    if (!guarded && infos->options->block && (infos->language==LANGUAGE_C))
    { strcat(seq,"{ ") ;
      *indent += INDENT_STEP ;
      *modulo_guard = 1 ;
    }
		
    sprintf(body,"%s = ",infos->iterators[level-1]) ;
    strcat(seq,body) ;
    sline = pprint_line(domain,*equal,num,level,infos) ;
    if (infos->language == LANGUAGE_FORTRAN)
    sprintf(body,"%s\n",sline) ;
    else
    sprintf(body,"%s ;\n",sline) ;
    strcat(seq,body) ;
    free(sline) ;
  }
  free(body) ;
  
  return(seq) ;
}
 

/* pprint_for function:
 * This function returns a string containing the printing of a loop header
 * according to an element.
 * A loop header according to an element is the conjonction of a minimum and a
 * maximum on the element (they give the loop bounds).
 * For instance, considering these constraints and the element j:
 * i + j -9*M >= 0
 *    -j +5*M >= 0
 *     j -4*M >= 0
 * this function should return 'for (j=max(-i+9*M,4*M),j<=5*M;j++) {'.
 * - domain is the polyhedron containing all the constraints,
 * - equal is a linked list containing the equalities to spread,
 * - level is the column number of the element in domain we want to use,
 * - accol is a pointer to a boolean set to 0 if there is no 'if', 1 otherwise,
 *   this function can set this boolean to 1 if there is a 'if',
 * - indent is a pointer to an integer giving the indent level, this function
 *   can increase indent if there is a 'if',
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * July 2nd 2002: first version (pick from pprint function). 
 * March 6th 2003: infinite domain support. 
 * June 29th 2003: non-unit strides support.
 */
char * pprint_for(domain, equal, level, accol, indent, infos)
Polyhedron * domain ;
CloogEqual ** equal ;
int level, * accol, * indent ;
CloogInfos * infos ;
{ int i ;
  char * sfor, * sline1, *sline2, * temp, * iterator ;
  
  iterator = infos->iterators[level-1] ;
  
  temp = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sfor = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sfor[0] = '\0' ;
        
  sline1 = pprint_minmax(domain,*equal,level,1,0,infos) ;
  sline2 = pprint_minmax(domain,*equal,level,0,0,infos) ;

  /* If min and max are not equal there is a 'for' else, there is a '='.
   * In the special case sline1 = sline2 = '\0', this is an infinite loop
   * so this is not a '='.
   */
  if ((strcmp(sline1,sline2) != 0) || !infos->options->otl ||
      ((strlen(sline1) == 0) && (strlen(sline2) == 0)))
  { for (i=0;i<(*indent);i++)
    strcat(sfor," ") ;
    
    if (infos->language == LANGUAGE_FORTRAN)
    sprintf(temp,"DO ") ;
    else
    sprintf(temp,"for (") ;
    
    strcat(sfor,temp) ;
    if (strlen(sline1) != 0)
    { sprintf(temp,"%s=%s",iterator,sline1) ;
      strcat(sfor,temp) ;
    }
    else
    if (infos->language == LANGUAGE_FORTRAN)
    { fprintf(stderr,"[CLooG]ERROR: unbounded loops not allowed in FORTRAN.\n");
      exit(1) ;
    }

    if (infos->language == LANGUAGE_FORTRAN)
    sprintf(temp,", ") ;
    else
    sprintf(temp,";") ;

    strcat(sfor,temp) ;
    if (strlen(sline2) != 0)
    { if (infos->language == LANGUAGE_FORTRAN)
      sprintf(temp,"%s",sline2) ;
      else
      sprintf(temp,"%s<=%s",iterator,sline2) ;
      
      strcat(sfor,temp) ;
    }
    else
    if (infos->language == LANGUAGE_FORTRAN)
    { fprintf(stderr,"[CLooG]ERROR: unbounded loops not allowed in FORTRAN.\n");
      exit(1) ;
    }
    
    if (infos->language == LANGUAGE_FORTRAN)
    { if (infos->stride[level-1] > 1)
      sprintf(temp,", "VALUE_FMT"\n",infos->stride[level-1]) ;
      else
      sprintf(temp,"\n") ;
    }
    else
    { if (infos->stride[level-1] > 1)
      sprintf(temp,";%s+="VALUE_FMT") {\n",iterator,infos->stride[level-1]) ;
      else
      sprintf(temp,";%s++) {\n",iterator) ;
    }
    
    *accol = 1 ;
    strcat(sfor,temp) ;
  }
  else
  if (!pprint_equal_add(equal,domain,level,ONE_TIME_LOOP,infos))
  { for (i=0;i<(*indent);i++)
    strcat(sfor," ") ;
    if (infos->language == LANGUAGE_FORTRAN)
    sprintf(temp,"%s = %s\n",iterator,sline1);
    else
    sprintf(temp,"%s = %s ;\n",iterator,sline1);
    strcat(sfor,temp) ;
  }
  free(temp) ;
  free(sline1) ;
  free(sline2) ;

  return(sfor) ;    
}


/* pprint_statement function:
 * This function returns a string containing the printing of a statement
 * structure.
 * - statement is the structure containing all the statement informations,
 * - equal is a linked list containing the equalities to spread,
 * - level is the number of loops enclosing the statement,
 * - accol is a pointer to a boolean set to 0 if there is no 'if' before the
 *   statement, 1 otherwise,
 * - indent is a pointer to an integer giving the indentation level,
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * September 21th 2003: first version (pick from pprint function). 
 */
char * pprint_statement(statement, equal, level, accol, indent, infos)
CloogStatement * statement ;
CloogEqual * equal ;
int level, accol, indent ;
CloogInfos * infos ;
{ int i, new_indent ;
  char * sstatement, * sline, * temp ;
 
  temp = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sstatement = (char *)malloc(MAX_STRING*sizeof(char)) ;
  sstatement[0] = '\0' ;
          
  while (statement != NULL)
  { new_indent = (accol == 1) ? indent + INDENT_STEP : indent ;
  
    /* Printing of one statement. */
    for (i=0;i<new_indent;i++)
    strcat(sstatement," ") ;

    if (infos->options->cpp == 0)
    sline = pprint_equal(equal,infos) ;
    else
    sline = pprint_equal_cpp(equal,level,infos) ;
    
    sprintf(temp,"%s%s",statement->body,sline) ;
    strcat(sstatement,temp) ;
    free(sline) ;     

    if (infos->language != LANGUAGE_FORTRAN)
    strcat(sstatement," ;") ;
    strcat(sstatement,"\n") ;
     
    statement = statement->next ;
  }
  free(temp) ;
  
  return(sstatement) ;
}


/* pprint function:
 * This function prints the content of a CloogLoop structure (loop) into a
 * file (foo, possibly stdout), in a C-like or FORTRAN-like language.
 * This function pretty-prints the data of a loop. The iterator (level) of
 * the current loop is given by 'level': this is the column number of the
 * domain corresponding to the current loop iterator. The data of a loop are
 * written in this order:
 * 1. The guard of the loop, i.e. each constraint in the domain that do not
 *    depend on the iterator (when the entry in the column 'level' is 0).
 * 2. The iteration domain of the iterator, given by the constraints in the
 *    domain depending on the iterator, i.e.:
 *    * an equality if the iterator has only one value (possibly preceded by
 *      a guard verifying if this value is integral), *OR*
 *    * a loop from the minimum possible value of the iterator to the maximum
 *      possible value.
 * 3. The included statements.
 * 4. The inner loops (recursive call).
 * 5. The following loops (recursive call).
 * - equal is a linked list containing the equalities to spread,
 * - level is the recursion level or the iteration level that we are printing,
 * - indent is the indent level, the number of spaces to print before writing,
 * - the infos structure gives the user some options about code printing,
 * - the infos structure gives the user some options about code printing,
 *   the number of parameters in domain (nb_par), and the arrays of iterator
 *   names and parameters (iters and params). 
 * November 2nd 2001: first version. 
 * March 6th 2003: infinite domain support. 
 * April 19th 2003: (debug) NULL loop support. 
 * June 29th 2003: non-unit strides support.
 */
void pprint(foo, loop, equal, level, indent, infos)
FILE * foo ;
CloogLoop * loop ;
CloogEqual * equal ;
int level, indent ;
CloogInfos * infos ;
{ int i, new_indent, stride, nb_equ=0, accol=0, guard=0, modulo_guard=0 ;
  char * sline ;
  Polyhedron * domain ;

  /* It can happen that loop be NULL when an input polyhedron is empty. */
  if (loop == NULL)
  return ;

  domain = loop->domain ;
  infos->stride[level-1] = loop->stride ;

  /* First of all we have to print the guard. */
  sline = pprint_guard(domain,equal,level,&guard,&indent,infos) ;
  fprintf(foo,"%s",sline) ;
  free(sline) ;
        
  if (domain->Dimension-infos->nb_parameters >= level)
  { /* We scan all the constraints to know in which case we are :
     * [[if] equality] or [for].
     */
    for (i=0;i<domain->NbConstraints;i++)
    if ((domain->Constraint[i][0] == 0) && (domain->Constraint[i][level] != 0))
    { /* If there is an equality, we can print it directly -no ambiguity-.
       * PolyLib can give more than one equality, we use just the first one
       * (this is a PolyLib problem, but all equalities are equivalent).
       */
      sline=pprint_equality(domain,&equal,i,level,&modulo_guard,&indent,infos) ;
      fprintf(foo,"%s",sline) ;
      free(sline) ;
      nb_equ = i + 1 ;   
      break ;
    }
  
    if (!nb_equ)
    { stride = 0 ;
      sline = pprint_for(domain,&equal,level,&accol,&indent,infos);
      fprintf(foo,"%s",sline) ;
      free(sline) ;
    }
  }
  
  /* Finally, if there is some included statements, print them. */
  sline = pprint_statement(loop->statement,equal,level,accol,indent,infos) ;
  fprintf(foo,"%s",sline) ;
  free(sline) ;

  /* Go to the next level. */
  if (loop->inner != NULL)
  { new_indent = (accol == 1) ? indent + INDENT_STEP : indent ;
    pprint(foo,loop->inner,equal,level+1,new_indent,infos) ;    
  }
  
  /* Close the open braces. */
  if (accol == 1)
  { for (i=0;i<indent;i++) fprintf(foo," ") ;
  
    if (infos->language == LANGUAGE_FORTRAN)
    fprintf(foo,"END DO\n") ; 
    else
    fprintf(foo,"}\n") ; 
  }
  if (guard == 1)
  { indent = indent - INDENT_STEP ;
    for (i=0;i<indent;i++) fprintf(foo," ") ;
    
    if (infos->language == LANGUAGE_FORTRAN)
    fprintf(foo,"END IF\n") ; 
    else
    fprintf(foo,"}\n") ; 
  }
  if (modulo_guard == 1)
  { indent = indent - INDENT_STEP ;
    for (i=0;i<indent;i++) fprintf(foo," ") ;
    
    if (infos->language == LANGUAGE_FORTRAN)
    fprintf(foo,"END IF\n") ; 
    else
    fprintf(foo,"}\n") ;
  }
  pprint_equal_del(&equal,level) ;
  /* Go to the next loop on the same level. */
  if (loop->next != NULL)
  pprint(foo,loop->next,equal,level,indent,infos) ;
}
