
   /**-------------------------------------------------------------------**
    **                              CLooG                                **
    **-------------------------------------------------------------------**
    **                           statement.c                             **
    **-------------------------------------------------------------------**
    **                 First version: november 4th 2001                  **
    **-------------------------------------------------------------------**/


/******************************************************************************
 *               CLooG : the Chunky Loop Generator (experimental)             *
 ******************************************************************************
 *                                                                            *
 * Copyright (C) 2001 Cedric Bastoul                                          *
 *                                                                            *
 * This is free software; you can redistribute it and/or modify it under the  *
 * terms of the GNU General Public License as published by the Free Software  *
 * Foundation; either version 2 of the License, or (at your option) any later *
 * version.                                                                   *
 *                                                                            *
 * This software is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License   *
 * for more details.                                                          *
 *                                                                            *
 * You should have received a copy of the GNU General Public License along    *
 * with software; if not, write to the Free Software Foundation, Inc.,        *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA                     *
 *                                                                            *
 * CLooG, the Chunky Loop Generator                                           *
 * Written by Cedric Bastoul, Cedric.Bastoul@inria.fr                         *
 *                                                                            *
 ******************************************************************************/
/* CAUTION: the english used for comments is probably the worst you ever read,
 *          please feel free to correct and improve it !
 */

# include <stdlib.h>
# include <stdio.h>
# include <string.h>
# include "../include/cloog/cloog.h"


/******************************************************************************
 *                             Memory leaks hunting                           *
 ******************************************************************************/


/* These functions and global variables are devoted to memory leaks hunting: we
 * want to know at each moment how many CloogStatement structures had been
 * allocated (cloog_statement_allocated) and how many had been freed
 * (cloog_statement_freed). Each time a CloogStatement structure is allocated,
 * a call to the function cloog_statement_leak_up() must be carried out, and
 * respectively cloog_statement_leak_down() when a CloogStatement structure is
 * freed. The special variable cloog_statement_max gives the maximal number of
 * CloogStatement structures simultaneously alive (i.e. allocated and
 * non-freed) in memory.
 * July 3rd->11th 2003: first version (memory leaks hunt and correction).
 */


int cloog_statement_allocated = 0 ;
int cloog_statement_freed = 0 ;
int cloog_statement_max = 0 ;


void cloog_statement_leak_up()
{ cloog_statement_allocated ++ ;
  if ((cloog_statement_allocated-cloog_statement_freed) > cloog_statement_max)
  cloog_statement_max = cloog_statement_allocated - cloog_statement_freed ;
}


void cloog_statement_leak_down()
{ cloog_statement_freed ++ ;
}


/******************************************************************************
 *                          Structure display function                        *
 ******************************************************************************/


/* cloog_statement_print function:
 * This function prints the content of a CloogStatement structure (statement)
 * into a file (foo, possibly stdout).
 */
void cloog_statement_print(FILE * foo, CloogStatement * statement)
{ if (statement != NULL)
  { fprintf(foo,"Included statements are:\n") ;
    
    while (statement != NULL)
    { fprintf(foo,"%s\n\n",statement->body) ;
      statement = statement->next ;
    }
  }
  else
  fprintf(foo,"No included statements.\n") ;  
}


/******************************************************************************
 *                         Memory deallocation function                       *
 ******************************************************************************/


/* cloog_statement_free function:
 * This function frees the allocated memory for a CloogStatement structure.
 */
void cloog_statement_free(CloogStatement * statement)
{ CloogStatement * next ;

  while (statement != NULL)
  { cloog_statement_leak_down() ;
    
    next = statement->next ;
    free(statement->body) ;
    /* free(statement->usr) ; Actually, this is user's job ! */
    free(statement) ;
    statement = next ;
  }
}


/******************************************************************************
 *                              Reading functions                             *
 ******************************************************************************/


/* cloog_statement_read function:
 * This function reads statement data from a file (foo) and puts them into
 * a CloogStatement structure. This function returns a pointer to this
 * structure.
 * - nb_iterators is the number of surrounding loops for the statement.
 * - number is the statement number.
 * September 9th 2002: first version.
 * March    17th 2003: fix for the usr field in CloogStatement structure.
 */
CloogStatement * cloog_statement_read(foo, nb_iterators, number)
FILE * foo ;
int nb_iterators, number ;
{ int op1, op2, op3 ;
  char s[MAX_STRING] ;
  CloogStatement * statement ;
  
  cloog_statement_leak_up() ;
  
  /* statement data. */
  statement = (CloogStatement *)malloc(sizeof(CloogStatement)) ;
  if (statement == NULL) 
  { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
    exit(1) ;
  }
  statement->body = (char *)malloc(sizeof(char)*MAX_STRING) ;
  if (statement->body == NULL) 
  { fprintf(stderr, "[CLooG]ERROR: memory overflow.\n") ;
    exit(1) ;
  }
  sprintf(statement->body, "S%d",number);
  
  /* statement->usr = (void *)malloc(sizeof(void)) ; Actually user's job ! */
  statement->usr  = NULL ;
  statement->next = NULL ;
  
  while (fgets(s,MAX_STRING,foo) == 0) ;
  while ((*s=='#'||*s=='\n') || (sscanf(s," %d",&op1)<1))
  fgets(s,MAX_STRING,foo) ;
  
  if (op1 != 0)
  { /* Do someting if the first option is set... */
    /* Nothing for the moment ! */
    
    /* Then read the other options. */
    while (fgets(s,MAX_STRING,foo) == 0) ;
    while ((*s=='#' || *s=='\n') || (sscanf(s," %d %d",&op2,&op3)<2))
    fgets(s,MAX_STRING,foo) ;
  }
  else
  { /* Do another thing if the first option is not set... */
    /* Nothing for the moment ! */
    
    /* Then read the other options. */
    if (sscanf(s," %d %d %d",&op1,&op2,&op3)<3)
    { while (fgets(s,MAX_STRING,foo) == 0) ;
      while ((*s=='#'||*s=='\n') || (sscanf(s," %d",&op1)<1))
      fgets(s,MAX_STRING,foo) ;
    }
  }
  
  return statement ;
}


/******************************************************************************
 *                            Processing functions                            *
 ******************************************************************************/


/* cloog_statement_copy function:
 * This function returns a copy of the CloogStatement structure given as input.
 * October 28th 2001: first version (in loop.c). 
 * March   17th 2003: fix for the usr field in CloogStatement structure.
 */ 
CloogStatement * cloog_statement_copy(CloogStatement * source)
{ CloogStatement * statement, * temp, * now = NULL ;
  
  statement = NULL ;

  while (source != NULL)
  { cloog_statement_leak_up() ;

    temp = (CloogStatement *)malloc(sizeof(CloogStatement)) ;
    if (temp == NULL)
    { fprintf(stderr, "Memory Overflow.\n") ;
      exit(1) ;
    }
    temp->body = (char *)malloc(sizeof(char)*MAX_STRING) ;
    if (temp->body == NULL)
    { fprintf(stderr, "Memory Overflow.\n") ;
      exit(1) ;
    }
    strcpy(temp->body,source->body) ;

    temp->usr     = source->usr ;
    temp->next    = NULL ;
    
    if (statement == NULL)
    { statement = temp ;
      now = statement ;
    }
    else
    { now->next = temp ;
      now = now->next ;
    }
    source = source->next ;
  }
  return(statement) ;
}


/* cloog_statement_add function:
 * This function adds a CloogStatement structure (statement) at a given place
 * (now) of a NULL terminated list of CloogStatement structures. The beginning
 * of this list is (start). This function updates (now) to (loop), and
 * updates (start) if the added element is the first one -that is when (start)
 * is NULL-.
 * March 27th 2004: first version. 
 */ 
void cloog_statement_add(start, now, statement)
CloogStatement ** start, ** now, * statement ;
{ if (*start == NULL)
  { *start = statement ;
    *now = *start ;
  }
  else
  { (*now)->next = statement ;
    *now = (*now)->next ;
  }
}

/* Function cloog_statement_begins:
 * This function returns 1 if the linked list S1 is the beginning of the
 * linked list S2, 0 otherwise.
 * March 27th 2004: first version. 
 */
int cloog_statement_begins(CloogStatement * S1, CloogStatement * S2)
{ while ((S1 != NULL) && (S2 != NULL))
  { if (strcmp(S1->body,S2->body) != 0)
    return 0 ;
    
    S1 = S1->next ;
    S2 = S2->next ;
  }
  if (S1 != NULL)
  return 0 ;
  
  return 1 ;
}


/* Function cloog_statement_begins:
 * This function returns 1 if the linked list S1 is the ending of the
 * linked list S2, 0 otherwise.
 * March 27th 2004: first version. 
 */
int cloog_statement_ends(CloogStatement * S1, CloogStatement * S2)
{ int i, nb_S1=0, nb_S2=0 ;
  CloogStatement * start_S1, * start_S2 ;
  
  start_S1 = S1 ;
  start_S2 = S2 ;
  /* We count the nember of element of each list (S2 must be >= S1). */
  while ((S1 != NULL) && (S2 != NULL))
  { nb_S1 ++ ;
    nb_S2 ++ ;
    S1 = S1->next ;
    S2 = S2->next ;
  }
  if (S1 != NULL)
  return 0 ;
  
  while (S2 != NULL)
  { nb_S2 ++ ;
    S2 = S2->next ;
  }
  
  S1 = start_S1 ;
  S2 = start_S2 ;
  
  /* We pass the first elements of S2 then we compare the ending. */
  for (i=0;i<nb_S2-nb_S1;i++)
  S2 = S2->next ;
  
  while ((S1 != NULL) && (S2 != NULL))
  { if (strcmp(S1->body,S2->body) != 0)
    return 0 ;
    
    S1 = S1->next ;
    S2 = S2->next ;
  }
  return 1 ;
}


/* Function cloog_statement_number:
 * This function returns the number of nodes in the linked list CloogStatement.
 * March 28th 2004: first version.
 */ 
int cloog_statement_number(CloogStatement * statement)
{ int nb_statements=0 ;

  while (statement != NULL)
  { nb_statements ++ ;
    statement = statement->next ;
  }
  return(nb_statements) ;
}













